#include <alsa/asoundlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <pthread.h>
#include <sys/select.h>
#include <sndfile.h>

#include "alsa.h"

typedef struct {
  snd_pcm_t *h;
  int frames_per_second;
  int channels;
  alsa_format_t f;
  int bytes_per_frame;
} alsa_data_t;

#define A do { if (err < 0) { fprintf(stderr, "%s:%d: alsa error '%s'\n", __FILE__, __LINE__, snd_strerror(err)); abort(); } } while (0)

void *open_alsa(int samplerate, int channels, alsa_format_t f, int output,
   char *device)
{
  alsa_data_t *a;
  snd_pcm_t *h;
  snd_pcm_hw_params_t *hw_params;
  snd_pcm_sw_params_t *sw_params;
  snd_pcm_uframes_t   period_frames;
  snd_pcm_uframes_t   buffer_frames;
  unsigned int        rate;
  snd_pcm_format_t    format;
  int err;

  switch (f) {
    case ALSA_S16:   format = SND_PCM_FORMAT_S16_LE;   break;
    case ALSA_FLOAT: format = SND_PCM_FORMAT_FLOAT_LE; break;
  }
  if (!output)
    { err = snd_pcm_open(&h, device, SND_PCM_STREAM_CAPTURE, 0); A; }
  else
    { err = snd_pcm_open(&h, device, SND_PCM_STREAM_PLAYBACK, 0); A; }
  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_sw_params_alloca(&sw_params);
  err = snd_pcm_hw_params_any(h, hw_params); A;
  err = snd_pcm_hw_params_set_access(h, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED); A;
  err = snd_pcm_hw_params_set_format(h, hw_params, format); A;
  err = snd_pcm_hw_params_set_channels(h, hw_params, channels); A;
  rate = samplerate;
  err = snd_pcm_hw_params_set_rate_near(h, hw_params, &rate, 0); A;
  if (rate != samplerate) {printf("bad rate\n"); abort();}
  period_frames = 1024;
  buffer_frames = 1024*3;
  err = snd_pcm_hw_params_set_period_size_near(h,hw_params,&period_frames,0);A;
  err = snd_pcm_hw_params_set_buffer_size_near(h, hw_params, &buffer_frames);A;
  err = snd_pcm_hw_params(h, hw_params); A;
  a = malloc(sizeof(*a)); if (a == NULL) abort();
  a->h = h;
  a->frames_per_second = samplerate;
  a->channels = channels;
  a->f = f;
  switch (f) {
    case ALSA_S16:   a->bytes_per_frame = 2; break;
    case ALSA_FLOAT: a->bytes_per_frame = 4; break;
  }
  a->bytes_per_frame *= channels;
  return a;
}

void close_alsa(void *_)
{
  alsa_data_t *a = _;
  snd_pcm_close(a->h);
  free(a);
}

static void xrun(snd_pcm_t *h)
{
  snd_pcm_prepare(h);
}

static void suspend(snd_pcm_t *h)
{
  int res;
  int loop = 0;
  while ((res = snd_pcm_resume(h)) == -EAGAIN && loop < 100) {
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10 * 1000;
    select(0, 0, 0, 0, &t);
    loop++;
  }
  if (res < 0) snd_pcm_prepare(h);
}

void put_alsa(void *_, void *buf, int frames, float vol)
{
  alsa_data_t *a = _;
  snd_pcm_t *h = a->h;
  int bytes_per_frame = a->bytes_per_frame;
  int channels = a->channels;
#define X do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); xrun(h); } while (0)
#define Y do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); suspend(h); } while (0)
  ssize_t r;
  unsigned char *bb;
  size_t count = frames;

  /* todo: speed up */
  switch (a->f) {
  case ALSA_S16: {
    short *b = buf;
    int i;
    for (i = 0; i < frames * channels; i++) {
      float v = b[i] * vol;
      if (v > 32767) v = 32767;
      if (v < -32768) v = -32768;
      b[i] = v;
    }
    break; }
  case ALSA_FLOAT: {
    float *b = buf;
    int i;
    for (i = 0; i < frames * channels; i++) b[i] *= vol;
    break; }
  }

  bb = buf;

  while (count) {
    r = snd_pcm_writei(h, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) { usleep(50*1000); X; }
    else if (r == -ESTRPIPE) Y;
    else if (r < 0) abort();
    if (r > 0) {
      count -= r;
      bb += r * bytes_per_frame;
    }
  }
#undef X
#undef Y
}

void get_alsa(void *_, void *buf, int frames)
{
  alsa_data_t *a = _;
  snd_pcm_t *h = a->h;
  int bytes_per_frame = a->bytes_per_frame;
#define X do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); xrun(h); } while (0)
#define Y do { fprintf(stderr, "%s:%s:%d: alsa error\n", __FILE__, __FUNCTION__, __LINE__); suspend(h); } while (0)
  ssize_t r;
  unsigned char *bb;
  size_t count = frames;

  bb = buf;

  while (count) {
    r = snd_pcm_readi(h, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) { usleep(50*1000); X; }
    else if (r == -ESTRPIPE) Y;
    else if (r < 0) abort();
    if (r > 0) {
      count -= r;
      bb += r * bytes_per_frame;
    }
  }
#undef X
#undef Y
}

uint64_t alsa_get_latency(void *_)
{
  alsa_data_t *a = _;
  snd_pcm_t *h = a->h;
  int frames_per_second = a->frames_per_second;
  int ret;
  snd_pcm_sframes_t fr = 0;

  ret = snd_pcm_delay(h, &fr);
  if (ret) fr = 0;
  /* useless test? can it be <0? */
  if (fr < 0) fr = 0;

  return (uint64_t)fr * (uint64_t)1000000 / (uint64_t)frames_per_second;
}
