#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <signal.h>
#include <pthread.h>

#include "utils.h"
#include "queue.h"
#include "alsa.h"
#include "volume.h"

/*
You can also do this using a unix socket to avoid some overhead of TCP:

/etc/pulse/default.pa:
load-module module-native-protocol-unix auth-anonymous=1 socket=/tmp/my-pulse-socket-name

/etc/pulse/client.conf:
default-server = unix:/tmp/my-pulse-socket-name
*/

#if 0
  tags:
    PA_TAG_INVALID = 0,
    PA_TAG_STRING = 't',
    PA_TAG_STRING_NULL = 'N',
    PA_TAG_U32 = 'L',
    PA_TAG_U8 = 'B',
    PA_TAG_U64 = 'R',
    PA_TAG_S64 = 'r',
    PA_TAG_SAMPLE_SPEC = 'a',
    PA_TAG_ARBITRARY = 'x',
    PA_TAG_BOOLEAN_TRUE = '1',
    PA_TAG_BOOLEAN_FALSE = '0',
    PA_TAG_BOOLEAN = PA_TAG_BOOLEAN_TRUE,
    PA_TAG_TIMEVAL = 'T',
    PA_TAG_USEC = 'U'  /* 64bit unsigned */,
    PA_TAG_CHANNEL_MAP = 'm',
    PA_TAG_CVOLUME = 'v',
    PA_TAG_PROPLIST = 'P',
    PA_TAG_VOLUME = 'V',
    PA_TAG_FORMAT_INFO = 'f',
#endif

/*
header:
    p->write.descriptor[PA_PSTREAM_DESCRIPTOR_LENGTH] = 0;
    p->write.descriptor[PA_PSTREAM_DESCRIPTOR_CHANNEL] = htonl((uint32_t) -1);
    p->write.descriptor[PA_PSTREAM_DESCRIPTOR_OFFSET_HI] = 0;
    p->write.descriptor[PA_PSTREAM_DESCRIPTOR_OFFSET_LO] = 0;
    p->write.descriptor[PA_PSTREAM_DESCRIPTOR_FLAGS] = 0;
see src/pulsecore/pstream.c

1st message: PA_COMMAND_AUTH (see setup_context())

header created by pa_tagstruct_command()
header: u32 = PA_COMMAND_AUTH (8)
        u32 = tag             (start from 0, increase?)
data:   u32 = PA_PROTOCOL_VERSION (32) [ | PA_PROTOCOL_FLAG_SHM (0x80000000) ]
                                       [ | PA_PROTOCOL_FLAG_MEMFD (0x40000000) ]
        arbitrary = cookie

u32 = 4c xx xx xx xx (big endian)
arbitrary = 78 sz sz sz sz xx .. xx (size is sz sz sz sz, big endian)

------------------------------------------------------------------------

reply = reply_new(tag)
    pa_tagstruct_putu32(reply, PA_PROTOCOL_VERSION | (do_shm ? 0x80000000 : 0) |
                        (do_memfd ? 0x40000000 : 0));

reply_new():
    reply = pa_tagstruct_new();
    pa_tagstruct_putu32(reply, PA_COMMAND_REPLY);
    pa_tagstruct_putu32(reply, tag);

reply message:
00 00 00 0f ff ff ff ff 00 00 00 00 00 00 00 00
00 00 00 00 4c 00 00 00 02 4c 00 00 00 00 4c 00
00 00 20 

*/

typedef struct {
  queue_t q;
  volatile int run;
  volatile int64_t read_pos;
  volatile int64_t write_pos;
  void *h;
  int channels;
  int samplerate;
  int bytes_per_sample;
  void *volume_module;
  int volume_id;
  int fd;
  int channel_id;
  pthread_mutex_t *fd_lock;
  char prebuf[4096];
  int prebuf_size;
} player_data_t;

void *play_thread(void *_)
{
  player_data_t *p = _;
  int free_size;
  buffer reply;

  while (p->run) {
    char b[4096];
    dequeue(&p->q, b);
    if (!p->run) break;
    p->read_pos += 4096;
    free_size = 16 - p->q.size;
    if (free_size < 0) free_size = 0;
    while (free_size) {
      int l;
      new_reply_more_data(&reply, p->channel_id, 4096);
      if (pthread_mutex_lock(p->fd_lock)) abort();
      l = fullwrite(p->fd, reply.buf, reply.len);
      if (pthread_mutex_unlock(p->fd_lock)) abort();
      if (l != reply.len) { printf("error writing\n"); }
      free_size--;
    }
    put_alsa(p->h, b, 4096 / p->bytes_per_sample,
             volume_db2lin(get_volume(p->volume_module, p->volume_id)));
  }

  close_alsa(p->h);
  free_queue(&p->q);
  free(p);
  return 0;
}

typedef struct {
  volatile int run;
  volatile int64_t read_pos;
  volatile int64_t write_pos;
  int channels;
  int samplerate;
  int bytes_per_sample;
  alsa_format_t format;
  int fd;
  int channel_id;
  pthread_mutex_t *fd_lock;
  char *alsa_in;
} recorder_data_t;

#include <arpa/inet.h>
void *record_thread(void *_)
{
  char header[4*5];
  char audio[4096];
  uint32_t *x;
  recorder_data_t *r = _;
  int l;
  void *h = open_alsa(r->samplerate, r->channels, r->format, 0, r->alsa_in);

  memset(header, 0, 20);
  for (int i = 0; i < 4096; i++) audio[i] = i;

  x = (uint32_t *)&header[0];
  *x = htonl(4096);

  x = (uint32_t *)&header[4];
  *x = htonl(r->channel_id);

  while (r->run) {
    if (!r->run) break;
    get_alsa(h, audio, 4096 / r->bytes_per_sample);
    r->read_pos += 4096;
    r->write_pos += 4096;
    if (pthread_mutex_lock(r->fd_lock)) abort();
    l = fullwrite(r->fd, header, 20);
    if (l != 20) goto over;
    l = fullwrite(r->fd, audio, 4096);
    if (l != 4096) goto over;
    if (pthread_mutex_unlock(r->fd_lock)) abort();
    //usleep(20*1000);
  }

over:
  close_alsa(h);
  while (r->run) usleep(100*1000);

  free(r);
  return 0;
}

typedef struct {
  int fd;
  void *volume_module;
  pthread_mutex_t lock;
  char *alsa_in;
  char *alsa_out;
} connection_data_t;

static void add_byte(player_data_t *p, char b)
{
  if (p->prebuf_size == 4096) {
    enqueue(&p->q, p->prebuf);
    p->prebuf_size = 0;
  }

  p->prebuf[p->prebuf_size] = b;
  p->prebuf_size++;
}

static void feed_data(player_data_t *p, char *buf, int len)
{
  int i;

  /* fast path */
  if (p->prebuf_size == 0 && len == 4096) {
    enqueue(&p->q, buf);
    return;
  }

  /* slow path, todo: optimize */
  for (i = 0; i < len; i++) add_byte(p, buf[i]);

  if (p->prebuf_size == 4096) {
    enqueue(&p->q, p->prebuf);
    p->prebuf_size = 0;
  }
}

void *connection_thread(void *_)
{
  connection_data_t *d = _;
  int fd2 = d->fd;
  player_data_t **plist   = NULL;
  int psize               = 0;
  int pmaxsize            = 0;
  recorder_data_t **rlist = NULL;
  int rsize               = 0;
  int rmaxsize            = 0;
  buffer reply;
  int i;
  int c;

#define WRITE() \
  do { \
    int l; \
    if (pthread_mutex_lock(&d->lock)) abort(); \
    l = fullwrite(fd2, reply.buf, reply.len); \
    if (pthread_mutex_unlock(&d->lock)) abort(); \
    if (l != reply.len) goto over; \
  } while (0)

  while (1) {
    message m;
    buffer r;

    read_message(fd2, &r, &m);

    if (m.length == -1) break;

    //printf("got message\n");
    if (m.channel != 0xffffffff) {
      //printf("data len %d\n", r.len);
      if (m.channel >= pmaxsize || plist[m.channel] == NULL) {
        printf("bad channel, won't ask for more data\n");
      } else {
        if (r.len > 0) {
          feed_data(plist[m.channel], (char*)r.buf, r.len);
          plist[m.channel]->write_pos += r.len;
          //new_reply_more_data(&reply, m.channel, 4096);
          //WRITE();
        } else {
          printf("bad data packet length %d, drop it\n", r.len);
        }
      }
      continue;
    }

    switch (m.type) {
    default: printf("unhandle message type %d\n", m.type); exit(1); break;

    case VOLUME_CLIENT:
      printf("switch to volume client mode\n");
      run_volume_client(d->volume_module, d->fd);
      goto over;

    case PA_COMMAND_AUTH: printf("got PA_COMMAND_AUTH\n");
      new_reply(&reply, -1, m.tag);
      putu32(&reply, 0x08 /* 0x20 */);   /* protocol version */
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_SET_CLIENT_NAME: printf("got PA_COMMAND_SET_CLIENT_NAME\n");
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_CREATE_PLAYBACK_STREAM: printf("got PA_COMMAND_CREATE_PLAYBACK_STREAM\n");
      alsa_format_t format;
      switch (m.c.playback.spec.format) {
      case 3: format = ALSA_S16;   break;
      case 5: format = ALSA_FLOAT; break;
      default:printf("unhandled format %d\n",m.c.playback.spec.format);exit(1);
      }
      if (psize == pmaxsize) {
        int inc = 64;
        pmaxsize += inc;
        plist = realloc(plist, pmaxsize * sizeof(*plist));
        if (plist == NULL) abort();
        for (i = pmaxsize - inc; i < pmaxsize; i++) plist[i] = NULL;
      }
      for (i = 0; i < pmaxsize; i++) if (plist[i] == NULL) break;
      new_reply(&reply, -1, m.tag);
      putu32(&reply, i);     /* s->index */
      putu32(&reply, i);     /* s->sink_input->index */
      putu32(&reply, 4096);  /* missing */
      size_reply(&reply);
      WRITE();
      psize++;
      plist[i] = malloc(sizeof(*plist[i])); if (plist[i] == NULL) abort();
      plist[i]->run = 1;
      plist[i]->read_pos = 0;
      plist[i]->write_pos = 0;
      plist[i]->channels   = m.c.playback.spec.channels;
      if (plist[i]->channels != 2 && plist[i]->channels != 1)
        {printf("%d channels not supported\n", plist[i]->channels); exit(1);}
      plist[i]->samplerate = m.c.playback.spec.rate;
      switch (m.c.playback.spec.format) {
      default: printf("bugbug!\n"); exit(1);
      case 3: plist[i]->bytes_per_sample = 2; break;
      case 5: plist[i]->bytes_per_sample = 4; break;
      }
      plist[i]->bytes_per_sample *= plist[i]->channels;
      plist[i]->h=open_alsa(plist[i]->samplerate,plist[i]->channels,format,1,
                            d->alsa_out);
      plist[i]->volume_module = d->volume_module;
      plist[i]->volume_id = new_volume(d->volume_module);
      plist[i]->fd = fd2;
      plist[i]->channel_id = i;
      plist[i]->fd_lock = &d->lock;
      plist[i]->prebuf_size = 0;
      init_queue(&plist[i]->q);
      new_thread(play_thread, plist[i]);
      break;

    case PA_COMMAND_DRAIN_PLAYBACK_STREAM: printf("got PA_COMMAND_DRAIN_PLAYBACK_STREAM\n");
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_DELETE_PLAYBACK_STREAM: printf("got PA_COMMAND_DELETE_PLAYBACK_STREAM\n");
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      c = m.c.delete_playback.index;
      if (c >= pmaxsize || plist[c]==NULL) {
        printf("bad channel\n");
        goto over;
      }
      free_volume(d->volume_module, plist[c]->volume_id);
      plist[c]->run = 0;
      kill_queue(&plist[c]->q);
      plist[c] = NULL;
      psize--;
      break;

    case PA_COMMAND_GET_SINK_INFO: printf("got PA_COMMAND_GET_SINK_INFO\n");
      goto sil;

    case PA_COMMAND_GET_SINK_INFO_LIST: printf("got PA_COMMAND_GET_SINK_INFO_LIST\n");
sil:
      new_reply(&reply, -1, m.tag);
      putu32(&reply, 0 /* index */);
      put_string(&reply, "sink");
      put_string(&reply, "sink output audio");
      put_sample_spec(&reply, 5, 2, 48000);
      put_channel_map(&reply, 2, (int[]){1, 2});
      putu32(&reply, 0 /* sink module index */);
      put_cvolume(&reply, 2, (int[]){0x10000, 0x10000});
      put_bool(&reply, 0 /* mute */);
      putu32(&reply, 0 /* source monitor index */);
      put_string(&reply, "sink.monitor");
      put_usec(&reply, 50*1000 /* latency */);
      put_string(&reply, "sink_driver.c");
      putu32(&reply, 0xe6 /* sink->flags & PA_SINK_CLIENT_FLAGS_MASK */);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_GET_SOURCE_INFO: printf("got PA_COMMAND_GET_SOURCE_INFO\n");
      goto sourcel;

    case PA_COMMAND_GET_SOURCE_INFO_LIST: printf("got PA_COMMAND_GET_SOURCE_INFO_LIST\n");
sourcel:
      new_reply(&reply, -1, m.tag);
      putu32(&reply, 0 /* index */);
      put_string(&reply, "source");
      put_string(&reply, "source input audio");
      put_sample_spec(&reply, 5, 1, 48000);
      put_channel_map(&reply, 1, (int[]){1});
      putu32(&reply, 0 /* source module index */);
      put_cvolume(&reply, 1, (int[]){0x10000});
      put_bool(&reply, 0 /* mute */);
      putu32(&reply, 0 /* source monitor index */);
      put_string(&reply, "source.monitor");
      put_usec(&reply, 50*1000 /* latency */);
      put_string(&reply, "source_driver.c");
      putu32(&reply, 0xe6 /* source->flags & PA_SOURCE_CLIENT_FLAGS_MASK */);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_SUBSCRIBE: printf("got PA_COMMAND_SUBSCRIBE (mask %d)\n", m.c.subscribe.mask);
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_GET_SERVER_INFO: printf("got PA_COMMAND_GET_SERVER_INFO\n");
      new_reply(&reply, -1, m.tag);
      put_string(&reply, "pulseaudionano");
      put_string(&reply, "1.0");
      put_string(&reply, "someone");
      put_string(&reply, "somewhere");
      put_sample_spec(&reply, 5 /* floatle */, 2, 48000);
      put_string(&reply, "sink");
      put_string(&reply, "source");
      putu32(&reply, 0 /* cookie */);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_GET_PLAYBACK_LATENCY: printf("got PA_COMMAND_GET_PLAYBACK_LATENCY\n");
      uint64_t latency;
      c = m.c.playback_latency.idx;
      if (c >= pmaxsize || plist[c]==NULL) {
        c = -1;
        printf("bad channel, send fake latency\n");
        latency = 50 * 1000; /* 50ms */
      } else {
        /* that works? two threads doing alsa stuff at the same time... */
        latency = alsa_get_latency(plist[c]->h);
        latency += (uint64_t)plist[c]->q.size *
                   (uint64_t)(4096/plist[c]->bytes_per_sample) *
                   (uint64_t)1000000 / (uint64_t)plist[c]->samplerate;
      }
printf("latency %ld id %d (%d bps %d samplerate qsize %d)\n", latency, c, plist[c]->bytes_per_sample, plist[c]->samplerate, plist[c]->q.size);

      new_reply(&reply, -1, m.tag);
      put_usec(&reply, latency);    /* current latency */
      put_usec(&reply, 0);          /* always 0 (?) */
      put_bool(&reply, 1);          /* playing? */
      put_timeval(&reply, m.c.playback_latency.time);
      put_timeval(&reply, get_time_now());
      /* this order works better for audio and video synchronization
       * in firefox
       * it should be the other way around (this or my brain is lost)
       */
      puts64(&reply, c == -1 ? 0 : plist[c]->read_pos);
      puts64(&reply, c == -1 ? 0 : plist[c]->write_pos);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_SET_SINK_INPUT_VOLUME: printf("got PA_COMMAND_SET_SINK_INPUT_VOLUME\n");
#if 0
{
  int i;
  printf("vol");
  for (i = 0; i < m.c.si_volume.volume.count; i++)
    printf(" 0x%x (%d)", m.c.si_volume.volume.volume[i], m.c.si_volume.volume.volume[i]);
  printf("\n");
/* -30 0
    y  x
    0  65536
  y/-30 = (x-65536)/-65536
  y = (x-65536) * 30 / 65536
 */
  set_volume(d->volume_module, 0,
             m.c.si_volume.volume.volume[0] ?
               ((int)m.c.si_volume.volume.volume[0] - 65536) * 30 / 65536 : -100);
}
#endif
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_CORK_PLAYBACK_STREAM: printf("got PA_COMMAND_CORK_PLAYBACK_STREAM %d %d\n", m.c.cork_playback.idx, m.c.cork_playback.bool);
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_TRIGGER_PLAYBACK_STREAM: printf("got PA_COMMAND_TRIGGER_PLAYBACK_STREAM\n");
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_CREATE_RECORD_STREAM: printf("got PA_COMMAND_CREATE_RECORD_STREAM\n");
      switch (m.c.record.spec.format) {
      case 3: format = ALSA_S16;   break;
      case 5: format = ALSA_FLOAT; break;
      default:printf("unhandled format %d\n",m.c.record.spec.format);exit(1);
      }
      if (rsize == rmaxsize) {
        int inc = 64;
        rmaxsize += inc;
        rlist = realloc(rlist, rmaxsize * sizeof(*rlist));
        if (rlist == NULL) abort();
        for (i = rmaxsize - inc; i < rmaxsize; i++) rlist[i] = NULL;
      }
      for (i = 0; i < rmaxsize; i++) if (rlist[i] == NULL) break;
      new_reply(&reply, -1, m.tag);
      putu32(&reply, i);     /* s->index */
      putu32(&reply, i);     /* s->source_output->index */
      size_reply(&reply);
      WRITE();
      rsize++;
      rlist[i] = malloc(sizeof(*rlist[i])); if (rlist[i] == NULL) abort();
      rlist[i]->run = 1;
      rlist[i]->read_pos = 0;
      rlist[i]->write_pos = 0;
      rlist[i]->channels   = m.c.record.spec.channels;
      if (rlist[i]->channels != 2 && rlist[i]->channels != 1)
        {printf("%d channels not supported\n", rlist[i]->channels); exit(1);}
      rlist[i]->samplerate = m.c.record.spec.rate;
      switch (m.c.record.spec.format) {
      default: printf("bugbug!\n"); exit(1);
      case 3: rlist[i]->bytes_per_sample = 2; break;
      case 5: rlist[i]->bytes_per_sample = 4; break;
      }
      rlist[i]->bytes_per_sample *= rlist[i]->channels;
      rlist[i]->format = format;
      rlist[i]->fd = fd2;
      rlist[i]->channel_id = i;
      rlist[i]->fd_lock = &d->lock;
      rlist[i]->alsa_in = d->alsa_in;
      new_thread(record_thread, rlist[i]);
      break;

    case PA_COMMAND_FLUSH_PLAYBACK_STREAM: printf("got PA_COMMAND_FLUSH_PLAYBACK_STREAM %d\n", m.c.flush_playback.idx);
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_GET_RECORD_LATENCY: printf("got PA_COMMAND_GET_RECORD_LATENCY\n");
      { uint64_t latency;
      c = m.c.record_latency.idx;
      if (c >= rmaxsize || rlist[c]==NULL) {
        c = -1;
        printf("bad channel, send fake latency\n");
        latency = 50 * 1000; /* 50ms */
      } else {
        /* todo (or not) */
        latency = 50 * 1000;
      }
      new_reply(&reply, -1, m.tag);
      put_usec(&reply, 0);          /* current monitor latency */
      put_usec(&reply, latency);    /* current source latency */
      put_bool(&reply, 1);          /* recording? */
      put_timeval(&reply, m.c.record_latency.time);
      put_timeval(&reply, get_time_now());
      puts64(&reply, c == -1 ? 0 : rlist[c]->write_pos);
      puts64(&reply, c == -1 ? 0 : rlist[c]->read_pos);
      size_reply(&reply);
      WRITE();
      break; }

    case PA_COMMAND_CORK_RECORD_STREAM: printf("got PA_COMMAND_CORK_RECORD_STREAM %d %d\n", m.c.cork_record.idx, m.c.cork_record.bool);
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_FLUSH_RECORD_STREAM: printf("got PA_COMMAND_FLUSH_RECORD_STREAM %d\n", m.c.flush_record.idx);
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      break;

    case PA_COMMAND_DELETE_RECORD_STREAM: printf("got PA_COMMAND_DELETE_RECORD_STREAM\n");
      new_reply(&reply, -1, m.tag);
      size_reply(&reply);
      WRITE();
      c = m.c.delete_record.index;
      if (c >= rmaxsize || rlist[c]==NULL) {
        printf("bad channel\n");
        goto over;
      }
      rlist[c]->run = 0;
      rlist[c] = NULL;
      rsize--;
      break;
    }
  }

over:
  for (i = 0; i < pmaxsize; i++)
    if (plist[i] != NULL) {
      free_volume(d->volume_module, plist[i]->volume_id);
      plist[i]->run = 0;
      kill_queue(&plist[i]->q);
    }
  for (i = 0; i < rmaxsize; i++)
    if (rlist[i] != NULL) rlist[i]->run = 0;
  printf("done\n");
  close(fd2);
  free(d);
  free(plist);
  free(rlist);
  return 0;
}

void usage(void)
{
  printf("options:\n");
  printf("    -do <device>\n");
  printf("        use <device> as ALSA output (default: 'default')\n");
  printf("        a good device to try is: 'plug:dmix:{CARD 0}'\n");
  printf("    -di <device>\n");
  printf("        use <device> as ALSA input (default: 'default')\n");
  printf("        a good device to try is: 'plug:dsnoop:{CARD 0}'\n");
  exit(1);
}

int main(int n, char **v)
{
  char *alsa_in = "default";
  char *alsa_out = "default";
  struct sockaddr_un addr;
  socklen_t addrlen;
  char path[128];
  int fd;
  int fd2;
  int i;
  void *volume_module = create_volume_module();

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-di") && i<=n-2) { alsa_in = v[++i]; continue; }
    if (!strcmp(v[i], "-do") && i<=n-2) { alsa_out = v[++i]; continue; }
    usage();
  }

  if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) exit(1);

  sprintf(path, "/tmp/pulseaudionano.socket");
  //sprintf(path, "/run/user/%d/pulse/native", getuid());

  fd=socket(AF_UNIX,SOCK_STREAM,0); if(fd==-1){perror("socket");return 1;}

  unlink(path);
  addr.sun_family = AF_UNIX;
  addr.sun_path[0] = 0;
  strcat(addr.sun_path, path);

  if (bind(fd, (struct sockaddr *)&addr, sizeof(addr)) == -1)
    { perror("bind"); return 1; }

  if (listen(fd, 5) == -1) { perror("listen"); return 1; }

  while (1) {
    connection_data_t *d;
    addrlen = sizeof(addr);
    fd2 = accept(fd, (struct sockaddr *)&addr, &addrlen);
    if (fd2 == -1) { perror("accept"); return 1; }
    d = malloc(sizeof(*d)); if (d == NULL) return 1;
    d->fd = fd2;
    d->volume_module = volume_module;
    d->alsa_in = alsa_in;
    d->alsa_out = alsa_out;
    if (pthread_mutex_init(&d->lock, NULL)) abort();
    new_thread(connection_thread, d);
  }

  delete_volume_module(volume_module);
  return 0;
}
