#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <setjmp.h>
#include <pthread.h>

#include "utils.h"

#define ERR(r) longjmp(*(jmp_buf *)r->x, 1)

int readu8(buffer *r)
{
  if (r->pos + 1 > r->len) ERR(r);
  r->pos++;
  return r->buf[r->pos-1];
}

uint32_t readu32(buffer *r)
{
  /* 4 bytes, big endian */
  if (r->pos + 4 > r->len) ERR(r);
  r->pos += 4;
  return (r->buf[r->pos-4] << 24) |
         (r->buf[r->pos-3] << 16) |
         (r->buf[r->pos-2] <<  8) |
          r->buf[r->pos-1];
}

int getbool(buffer *r)
{
  int b = readu8(r);

  switch(b) {
  case 0x30: return 0;
  case 0x31: return 1;
  default: ERR(r);
  }
}

uint32_t getu32(buffer *r)
{
  if (readu8(r) != 0x4c) ERR(r);

  return readu32(r);
}

void get_arbitrary(buffer *r, char *buf, int *bufsize)
{
  uint32_t size;

  if (readu8(r) != 0x78) ERR(r);

  size = readu32(r);
  if (size > IN_MAX) ERR(r);

  /* then size bytes */
  if (r->pos + size > r->len) ERR(r);
  memcpy(buf, r->buf + r->pos, size);
  r->pos += size;
  *bufsize = size;
}

void get_string(buffer *r, char *s)
{
  int t = readu8(r);

  /* null? */
  if (t == 0x4e) { *s = 0; return; }

  if (t != 0x74) ERR(r);

  /* data up to and including terminal 0 */
  while (r->pos < r->len) {
    /* no need to check buffer overflow, r->len cannot be > IN_MAX */
    *s = r->buf[r->pos];
    r->pos++;
    if (*s == 0) return;
    s++;
  }

  ERR(r);
}

void get_sample_spec(buffer *r, sample_spec_t *s)
{
  if (readu8(r) != 0x61) ERR(r);

  s->format   = readu8(r);
  s->channels = readu8(r);
  s->rate     = readu32(r);
}

void get_channel_map(buffer *r, channel_map_t *c)
{
  int i;

  if (readu8(r) != 0x6d) ERR(r);

  c->count = readu8(r);
  if (c->count > CHANNEL_MAX) ERR(r);
  for (i = 0; i < c->count; i++) c->map[i] = readu8(r);
}

void get_cvolume(buffer *r, cvolume_t *v)
{
  int i;

  if (readu8(r) != 0x76) ERR(r);

  v->count = readu8(r);
  if (v->count > CHANNEL_MAX) ERR(r);
  for (i = 0; i < v->count; i++) v->volume[i] = readu32(r);
}

struct timeval get_timeval(buffer *r)
{
  struct timeval ret;
  uint32_t v;

  if (readu8(r) != 0x54) ERR(r);

  v = readu32(r);
  ret.tv_sec = v;
  v = readu32(r);
  ret.tv_usec = v;

  return ret;
}

void _decode_command(buffer *r, message *m)
{
  uint32_t command = getu32(r);
  uint32_t tag     = getu32(r);
//printf("command %d tag %d\n", command, tag);

  m->type = command;
  m->tag  = tag;

  switch (command) {
  default: printf("XXXXXXXXXXXXXXXXXXXXX:                                unhandled command %d\n", command); ERR(r);
  case PA_COMMAND_AUTH:
    m->c.auth.protocol_version = getu32(r);
    get_arbitrary(r, m->c.auth.cookie, &m->c.auth.cookie_size);
    if (m->c.auth.cookie_size != 256) ERR(r);
    break;
  case PA_COMMAND_SET_CLIENT_NAME:
    /* we don't care */
    break;
  case PA_COMMAND_REPLY:
    /* can be anything... that's a joke. */
    break;
  case PA_COMMAND_CREATE_PLAYBACK_STREAM:
    get_string(r, m->c.playback.media_name);
    get_sample_spec(r, &m->c.playback.spec);
    get_channel_map(r, &m->c.playback.map);
    { uint32_t v = getu32(r); if (v!=0xffffffff){printf("bad\n");exit(1);}}
    get_string(r, m->c.playback.dev);
    m->c.playback.maxlength = getu32(r);
    m->c.playback.corked = getbool(r);
    m->c.playback.tlength = getu32(r);
    m->c.playback.prebuf = getu32(r);
    m->c.playback.minreq = getu32(r);
    m->c.playback.syncid = getu32(r);
    get_cvolume(r, &m->c.playback.volume);
printf("media name '%s'\n", m->c.playback.media_name);
printf("format %d channels %d rate %d\n", m->c.playback.spec.format, m->c.playback.spec.channels, m->c.playback.spec.rate);
{int i;printf("channels [%d]:", m->c.playback.map.count);
  for(i=0;i<m->c.playback.map.count;i++)printf(" %d", m->c.playback.map.map[i]);
  printf("\n");}
printf("dev '%s' maxlength %d corked %d tlength %d prebuf %d minreq %d syncid %d\n", m->c.playback.dev, m->c.playback.maxlength, m->c.playback.corked, m->c.playback.tlength, m->c.playback.prebuf, m->c.playback.minreq, m->c.playback.syncid);
{int i;printf("volume [%d]:", m->c.playback.volume.count);
  for(i=0;i<m->c.playback.volume.count;i++)printf(" %d", m->c.playback.volume.volume[i]);
  printf("\n");}
    break;
  case PA_COMMAND_REQUEST:
    m->c.request.index = getu32(r);
    m->c.request.length = getu32(r);
    break;
  case PA_COMMAND_DRAIN_PLAYBACK_STREAM:
    m->c.drain.index = getu32(r);
    break;
  case PA_COMMAND_DELETE_PLAYBACK_STREAM:
    m->c.delete_playback.index = getu32(r);
    break;
  case PA_COMMAND_GET_SINK_INFO_LIST:
    /* nothing */
    break;
  case PA_COMMAND_GET_SOURCE_INFO_LIST:
    /* nothing */
    break;
  case PA_COMMAND_SUBSCRIBE:
    m->c.subscribe.mask = getu32(r);
    break;
  case PA_COMMAND_GET_SERVER_INFO:
    /* nothing */
    break;
  case PA_COMMAND_GET_PLAYBACK_LATENCY:
    m->c.playback_latency.idx = getu32(r);
    m->c.playback_latency.time = get_timeval(r);
    break;
  case PA_COMMAND_GET_SINK_INFO:
    m->c.sink_info.idx = getu32(r);
    get_string(r, m->c.sink_info.name);
    break;
  case PA_COMMAND_SET_SINK_INPUT_VOLUME:
    m->c.si_volume.idx = getu32(r);
    get_cvolume(r, &m->c.si_volume.volume);
    break;
  case PA_COMMAND_CORK_PLAYBACK_STREAM:
    m->c.cork_playback.idx = getu32(r);
    m->c.cork_playback.bool = getbool(r);
    break;
  case PA_COMMAND_TRIGGER_PLAYBACK_STREAM:
    m->c.trigger.idx = getu32(r);
    break;
  case PA_COMMAND_GET_SOURCE_INFO:
    m->c.source_info.idx = getu32(r);
    get_string(r, m->c.source_info.name);
    break;
  case PA_COMMAND_CREATE_RECORD_STREAM:
    get_string(r, m->c.record.name);
    get_sample_spec(r, &m->c.record.spec);
    get_channel_map(r, &m->c.record.map);
    m->c.record.source_index = getu32(r);
    get_string(r, m->c.record.source_name);
    m->c.record.maxlength = getu32(r);
    m->c.record.corked = getbool(r);
    m->c.record.fragsize = getu32(r);
printf("name '%s' source_name '%s'\n", m->c.record.name, m->c.record.source_name);
printf("format %d channels %d rate %d\n", m->c.record.spec.format, m->c.record.spec.channels, m->c.record.spec.rate);
{int i;printf("channels [%d]:", m->c.record.map.count);
  for(i=0;i<m->c.record.map.count;i++)printf(" %d", m->c.record.map.map[i]);
  printf("\n");}
printf("maxlength %d corked %d fragsize %d\n", m->c.record.maxlength, m->c.record.corked, m->c.record.fragsize);
    break;
  case PA_COMMAND_FLUSH_PLAYBACK_STREAM:
    m->c.flush_playback.idx = getu32(r);
    break;
  case PA_COMMAND_GET_RECORD_LATENCY:
    m->c.record_latency.idx = getu32(r);
    m->c.record_latency.time = get_timeval(r);
    break;
  case PA_COMMAND_CORK_RECORD_STREAM:
    m->c.cork_record.idx = getu32(r);
    m->c.cork_record.bool = getbool(r);
    break;
  case PA_COMMAND_FLUSH_RECORD_STREAM:
    m->c.flush_record.idx = getu32(r);
    break;
  case PA_COMMAND_DELETE_RECORD_STREAM:
    m->c.delete_record.index = getu32(r);
    break;
  }
}

/* returns 0 if ok, -1 if error */
int decode_command(buffer *r, message *m)
{
  int ret;

  r->pos = 0;
  r->x = malloc(sizeof(sigjmp_buf)); if (r->x == NULL) abort();

  /* setjmp magic */
  if (setjmp(*(jmp_buf *)r->x) == 0) {
    _decode_command(r, m);
    ret = 0;
  } else
    ret = -1;

  free(r->x);

  return ret;
}

int fullread(int fd, void *_buf, int count)
{
//printf("reading %d bytes\n", count);
  char *buf = _buf;
  int ret = 0;
  int l;
  while (count) {
    l = read(fd, buf, count);
    if (l <= 0) return -1;
    count -= l;
    buf += l;
    ret += l;
  }
  return ret;
}

int fullwrite(int fd, void *_buf, int count)
{
  char *buf = _buf;
  int ret = 0;
  int l;
  while (count) {
    l = write(fd, buf, count);
    if (l <= 0) return -1;
    count -= l;
    buf += l;
    ret += l;
  }
  return ret;
}

void dump(buffer *r, message *m)
{
return;
  int i;

  printf("%8.8x %8.8x %8.8x %8.8x %8.8x\n",
         m->length, m->channel, m->offset_hi, m->offset_lo, m->flags);

  for (i = 0; i < r->len; i++) {
    printf("%2.2x", r->buf[i]);
    if (i != r->len-1) printf(i % 16 == 15 ? "\n" : " ");
  }
  printf("\n");
}

/* returns m->length = -1 on error */
void read_message(int fd, buffer *r, message *m)
{
  uint32_t a, b, c, d, e;

  m->type = -1;

  /* get header */
  if (fullread(fd, &a, 4) != 4 ||
      fullread(fd, &b, 4) != 4 ||
      fullread(fd, &c, 4) != 4 ||
      fullread(fd, &d, 4) != 4 ||
      fullread(fd, &e, 4) != 4) goto err;

  m->length    = ntohl(a);
  m->channel   = ntohl(b);
  m->offset_hi = ntohl(c);
  m->offset_lo = ntohl(d);
  m->flags     = ntohl(e);

  if (m->length    == 0x12345678 &&
      m->channel   == 0xffffffff &&
      m->offset_hi == 0x34567890 &&
      m->offset_lo == 0x45678901 &&
      m->flags     == 0x56789012) {
    m->type = VOLUME_CLIENT;
    return;
  }

  if (m->length > IN_MAX)
    {printf("increase IN_MAX (is %d, wanted %d)\n",IN_MAX,m->length);exit(1);}

  /* get data */
  r->len = m->length;
  if (fullread(fd, r->buf, r->len) != r->len) goto err;

  dump(r, m);

  /* decode command if channel is -1 (correct?) */
  if (m->channel == 0xffffffff) if (decode_command(r, m) == -1) goto err;

  return;

err:
  printf("error read_message\n");
  m->length = -1;
}

void new_thread(void *(*f)(void *), void *data)
{
  pthread_t t;
  pthread_attr_t att;

  if (pthread_attr_init(&att))
    { fprintf(stderr, "pthread_attr_init err\n"); exit(1); }
  if (pthread_attr_setdetachstate(&att, PTHREAD_CREATE_DETACHED))
    { fprintf(stderr, "pthread_attr_setdetachstate err\n"); exit(1); }
  if (pthread_attr_setstacksize(&att, 10000000))
    { fprintf(stderr, "pthread_attr_setstacksize err\n"); exit(1); }
  if (pthread_create(&t, &att, f, data))
    { fprintf(stderr, "pthread_create err\n"); exit(1); }
  if (pthread_attr_destroy(&att))
    { fprintf(stderr, "pthread_attr_destroy err\n"); exit(1); }
}

#define FULL do{printf("%s:%d: full\n",__FILE__,__LINE__);abort();}while(0)

void writeu8(buffer *b, int v)
{
  if (b->len > IN_MAX-1) FULL;
  b->buf[b->len] = v;
  b->len++;
}

void writeu32(buffer *b, uint32_t v)
{
  if (b->len > IN_MAX-4) FULL;
  b->buf[b->len]   = (v >> 24) & 255;
  b->buf[b->len+1] = (v >> 16) & 255;
  b->buf[b->len+2] = (v >>  8) & 255;
  b->buf[b->len+3] =  v        & 255;
  b->len += 4;
}

void writeu64(buffer *b, uint64_t v)
{
  if (b->len > IN_MAX-8) FULL;
  b->buf[b->len]   = (v >> (32+24)) & 255;
  b->buf[b->len+1] = (v >> (32+16)) & 255;
  b->buf[b->len+2] = (v >> (32+ 8)) & 255;
  b->buf[b->len+3] = (v >>  32    ) & 255;
  b->buf[b->len+4] = (v >>     24) & 255;
  b->buf[b->len+5] = (v >>     16) & 255;
  b->buf[b->len+6] = (v >>      8) & 255;
  b->buf[b->len+7] =  v            & 255;
  b->len += 8;
}

void putu32(buffer *b, uint32_t v)
{
  writeu8(b, 0x4c);
  writeu32(b, v);
}

void puts64(buffer *b, int64_t v)
{
  writeu8(b, 0x72);
  writeu64(b, v);
}

void put_bool(buffer *b, int v)
{
  writeu8(b, v ? 0x31 : 0x30);
}

void put_usec(buffer *b, uint64_t v)
{
  writeu8(b, 0x55);
  writeu64(b, v);
}

void put_timeval(buffer *b, struct timeval v)
{
  uint32_t vv;
  writeu8(b, 0x54);
  vv = v.tv_sec;
  writeu32(b, vv);
  vv = v.tv_usec;
  writeu32(b, vv);
}

void put_string(buffer *b, char *s)
{
  writeu8(b, 0x74);
  while (*s) writeu8(b, *s++);
  writeu8(b, 0);
}

void put_sample_spec(buffer *b, int format, int channels, int samplerate)
{
  uint32_t sr = samplerate;
  writeu8(b, 0x61);
  writeu8(b, format);
  writeu8(b, channels);
  writeu32(b, sr);
}

void put_channel_map(buffer *b, int n, int *channels)
{
  int i;
  writeu8(b, 0x6d);
  writeu8(b, n);
  for (i = 0; i < n; i++) writeu8(b, channels[i]);
}

void put_cvolume(buffer *b, int n, int *volumes)
{
  int i;
  writeu8(b, 0x76);
  writeu8(b, n);
  for (i = 0; i < n; i++) { uint32_t v = volumes[i]; writeu32(b, v); }
}

void new_reply(buffer *b, int channel, int tag)
{
  b->len = 0;
  writeu32(b, 0);        /* size - set later */
  writeu32(b, channel);
  writeu32(b, 0);        /* offset hi */
  writeu32(b, 0);        /* offset lo */
  writeu32(b, 0);        /* flags */

  putu32(b, PA_COMMAND_REPLY);
  putu32(b, tag);
}

void size_reply(buffer *b)
{
  uint32_t v = b->len - 5*4;
  b->buf[0] = (v >> 24) & 255;
  b->buf[1] = (v >> 16) & 255;
  b->buf[2] = (v >>  8) & 255;
  b->buf[3] =  v        & 255;
}

void new_reply_more_data(buffer *b, int channel, int size)
{
  b->len = 0;
  writeu32(b, 0);          /* size - set later */
  writeu32(b, 0xffffffff); /* channel - has to be this */
  writeu32(b, 0);          /* offset hi */
  writeu32(b, 0);          /* offset lo */
  writeu32(b, 0);          /* flags */
  putu32(b, 0x3d);
  putu32(b, 0xffffffff);
  putu32(b, channel);
  putu32(b, size);
  size_reply(b);
}

struct timeval get_time_now(void)
{
  struct timeval t;
  gettimeofday(&t, NULL);
  return t;
}
